""" LCR-Net: Localization-Classification-Regression for Human Pose
Copyright (C) 2017 Gregory Rogez & Philippe Weinzaepfel

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>"""

from __future__ import print_function
import os, sys
import numpy as np

faster_path = os.path.join( os.path.dirname(__file__), 'py-faster-rcnn')
if not os.path.isdir(faster_path): 
    print("ERROR: download and install py-faster-rcnn and put a softlink in {:s}".format(faster_path))
sys.path.append( os.path.join( faster_path, 'tools') )
import _init_paths 
import caffe
from scipy.io import savemat
import h5py 
import cv2

NJOINTS = 13 # NUMBER OF JOINTS
NT = 5       # 2D + 3D 

pixel_mean = np.array([[[102.9801, 115.9465, 122.7717]]], dtype=np.float32)

def image_to_blob(im, scale, maxsize, pixelmean):
    imblob = im.astype(np.float32, copy=True) - pixelmean
    im_shape = imblob.shape
    im_size_min = np.min(im_shape[0:2])
    im_size_max = np.max(im_shape[0:2])
    im_scale = float(scale) / float(im_size_min)
    if np.round(im_scale * im_size_max) > maxsize:
        im_scale = float(maxsize) / float(im_size_max)
    imblob = cv2.resize(imblob, None, None, fx=im_scale, fy=im_scale, interpolation=cv2.INTER_LINEAR)
    return np.transpose(imblob, (2,0,1))[None,:,:,:], im_scale

def detect_pose( img_output_list, prototxt, caffemodel, anchor_poses, gpuid=-1):
    """
    detect poses in a list of image
    img_output_list: list of couple (path_to_image, path_to_outputfile)
    prototxt: file containing the deploy.prototxt
    caffemodel: file containing the caffemodel
    anchor_poses: file containing the anchor_poses or directly the anchor poses
    gpuid: -1 for using cpu mode, otherwise device_id
    """
    
    # check that the files exist 
    if not os.path.isfile(prototxt): print("prototxt file does not exist: {:s}".format(prototxt))
    if not os.path.isfile(caffemodel): print("caffemodel file does not exist: {:s}".format(caffemodel))
    if isinstance(anchor_poses,str) and not os.path.isfile(anchor_poses): print("anchor_poses file does not exist: {:s}".format(anchor_poses))
    
    # load the anchor poses and the network
    if isinstance(anchor_poses,str): anchor_poses = np.asarray( h5py.File( anchor_poses )['data'] ).astype(np.float32)
    if gpuid<0:
        caffe.set_mode_cpu()
    else:
        caffe.set_mode_gpu()
        caffe.set_device(gpuid)    
    net = caffe.Net(prototxt, caffemodel, caffe.TEST)
    
    output = []
    # iterate over image 
    for imgname, outputname in img_output_list:
        # load the images and prepare the blob
        im = cv2.imread( imgname )        
        im_blob, im_scale = image_to_blob(im, scale=600, maxsize=1000, pixelmean=pixel_mean)
        
        im_info = np.array([[im_blob.shape[2], im_blob.shape[3], im_scale]], dtype=np.float32)
        
        # perform forword 
        net.blobs['data'].reshape(*(im_blob.shape))
        net.blobs['im_info'].reshape(*(im_info.shape))
        blobs_out = net.forward(data=im_blob, im_info=im_info)
        
        # parse output 
        boxes = net.blobs['rois'].data[:,1:5] / im_scale
        scores = blobs_out['cls_prob']
        # project poses on boxes
        boxes_size = boxes[:,2:4]-boxes[:,0:2]
        offset = np.concatenate( ( boxes[:,:2], np.zeros((boxes.shape[0],3),dtype=np.float32)), axis=1) # x,y top-left corner for each box
        scale = np.concatenate( ( boxes_size[:,:2], np.ones((boxes.shape[0],3),dtype=np.float32)), axis=1) # width, height for each box
        offset_poses = np.tile( np.concatenate( [np.tile( offset[:,k:k+1], (1,NJOINTS)) for k in xrange(5)], axis=1), (1,anchor_poses.shape[0])) # x,y top-left corner for each pose 
        scale_poses = np.tile( np.concatenate( [np.tile( scale[:,k:k+1], (1,NJOINTS)) for k in xrange(5)], axis=1), (1,anchor_poses.shape[0]))
            # x- y- scale for each pose 
        pred_poses = offset_poses + np.tile( anchor_poses.reshape(1,-1), (boxes.shape[0],1) ) * scale_poses # put anchor poses into the boxes
        pred_poses += scale_poses * blobs_out['pose_pred'][:,NJOINTS*5:] # apply regression (do not consider the one for the background class)
        
        # we save only the poses with score over th with at minimum 500 ones 
        th = 0.1/(scores.shape[1]-1)
        Nmin = min(500, scores[:,1:].size-1)
        if np.sum( scores[:,1:]>th ) < Nmin: # set thresholds to keep at least Nmin boxes
            th = - np.sort( -scores[:,1:].ravel() )[Nmin-1]
        where = zip(*np.where(scores[:,1:]>=th )) # which one to save 
        nPP = len(where) # number to save 
        regpose2d = np.empty((nPP,NJOINTS*2), dtype=np.float32) # regressed 2D pose 
        regpose3d = np.empty((nPP,NJOINTS*3), dtype=np.float32) # regressed 3D pose 
        regscore = np.empty((nPP,1), dtype=np.float32) # score of the regressed pose 
        regprop = np.empty((nPP,1), dtype=np.float32) # index of the proposal among the candidate boxes 
        regclass = np.empty((nPP,1), dtype=np.float32) # index of the anchor pose class 
        for ii, (i,j) in enumerate(where):
            regpose2d[ii,:] = pred_poses[i, j*NJOINTS*5:j*NJOINTS*5+NJOINTS*2]
            regpose3d[ii,:] = pred_poses[i, j*NJOINTS*5+NJOINTS*2:j*NJOINTS*5+NJOINTS*5]
            regscore[ii,0] = scores[i,1+j]
            regprop[ii,0] = i+1
            regclass[ii,0] = j+1
        tosave = {'regpose2d': regpose2d, 
                  'regpose3d': regpose3d, 
                  'regscore': regscore, 
                  'regprop': regprop, 
                  'regclass': regclass, 
                  'rois': boxes,
                 }
        output.append( tosave )
        if outputname is not None:
          os.system('mkdir -p '+os.path.dirname(outputname))
          savemat(outputname, tosave, do_compression=True)
    return output
        
if __name__=='__main__':
    def usae():
        print("python detect_pose.py <imgname> <outputname> <prototxt> <caffemodel> <anchor_poses> <gpu_id>")
    if len(sys.argv)==1:
        usage()
    elif len(sys.argv)<6 or len(sys.argv)>=8:
        print("Invalid number of arguments")
        usage()
    else:    
        imgname = sys.argv[1]
        outputname = sys.argv[2]
        prototxt = sys.argv[3]
        caffemodel = sys.argv[4]
        anchor_poses = sys.argv[5]
        gpu = -1 if len(sys.argv)<7 else int(sys.argv[6])
        detect_pose( [(imgname,outputname)], prototxt, caffemodel, anchor_poses, gpu=gpu)
